@@ -1,6 +1,7 @@
 from abc import ABC, abstractmethod
 from contextlib import contextmanager
 import re
+from time import sleep
 from typing import Iterator, Optional, Tuple
 from urllib.parse import unquote as urlunquote
 
@@ -8,8 +9,8 @@ from pydantic import AnyHttpUrl, BaseModel, parse_obj_as, validator
 import requests
 
 from . import get_logger
-from .consts import known_instances
-from .dandiapi import DandiAPIClient, RemoteAsset, RemoteDandiset
+from .consts import VERSION_REGEX, known_instances
+from .dandiapi import BaseRemoteAsset, DandiAPIClient, RemoteDandiset
 from .exceptions import FailedToConnectError, NotFoundError, UnknownURLError
 from .utils import get_instance
 
@@ -30,7 +31,7 @@ class ParsedDandiURL(ABC, BaseModel):
     #: The base URL of the Dandi API service, without a trailing slash
     api_url: AnyHttpUrl
     #: The ID of the Dandiset given in the URL
-    dandiset_id: str
+    dandiset_id: Optional[str]
     #: The version of the Dandiset, if specified.  If this is not set, methods
     #: that need the Dandiset version will call `get_version_id()` to get an
     #: appropriate default value.
@@ -44,12 +45,15 @@ class ParsedDandiURL(ABC, BaseModel):
         """Returns an unauthenticated `DandiAPIClient` for `api_url`"""
         return DandiAPIClient(self.api_url)
 
-    def get_dandiset(self, client: DandiAPIClient) -> RemoteDandiset:
+    def get_dandiset(self, client: DandiAPIClient) -> Optional[RemoteDandiset]:
         """
         Returns information about the specified (or default) version of the
         specified Dandiset
         """
-        return client.get_dandiset(self.dandiset_id, self.get_version_id(client))
+        if self.dandiset_id is not None:
+            return client.get_dandiset(self.dandiset_id, self.get_version_id(client))
+        else:
+            return None
 
     def get_version_id(self, client: DandiAPIClient) -> str:
         """
@@ -59,7 +63,9 @@ class ParsedDandiURL(ABC, BaseModel):
         most recent published version of the Dandiset is returned, if any,
         otherwise the ID of the draft version is returned.
         """
-        if self.version_id is None:
+        if self.dandiset_id is None:
+            raise ValueError("Cannot get version for Dandiset-less URL")
+        elif self.version_id is None:
             r = client.get(f"/dandisets/{self.dandiset_id}/")
             version = r["most_recent_published_version"] or r["draft_version"]
             return version["version"]
@@ -67,7 +73,7 @@ class ParsedDandiURL(ABC, BaseModel):
             return self.version_id
 
     @abstractmethod
-    def get_assets(self, client: DandiAPIClient) -> Iterator[RemoteAsset]:
+    def get_assets(self, client: DandiAPIClient) -> Iterator[BaseRemoteAsset]:
         """
         Returns an iterator of asset structures for the assets referred to by
         or associated with the URL.  For a URL pointing to just a Dandiset,
@@ -88,7 +94,9 @@ class ParsedDandiURL(ABC, BaseModel):
     @contextmanager
     def navigate(
         self,
-    ) -> Iterator[Tuple[DandiAPIClient, RemoteDandiset, Iterator[RemoteAsset]]]:
+    ) -> Iterator[
+        Tuple[DandiAPIClient, Optional[RemoteDandiset], Iterator[BaseRemoteAsset]]
+    ]:
         """
         A context manager that returns a triple of a `DandiAPIClient` (with an
         open session that is closed when the context manager closes), the
@@ -106,7 +114,7 @@ class DandisetURL(ParsedDandiURL):
     Parsed from a URL that only refers to a Dandiset (possibly with a version)
     """
 
-    def get_assets(self, client: DandiAPIClient) -> Iterator[RemoteAsset]:
+    def get_assets(self, client: DandiAPIClient) -> Iterator[BaseRemoteAsset]:
         """Returns all assets in the Dandiset"""
         return self.get_dandiset(client).get_assets()
 
@@ -123,19 +131,47 @@ class MultiAssetURL(ParsedDandiURL):
     path: str
 
 
-class AssetIDURL(SingleAssetURL):
-    """Parsed from a URL that refers to an asset by ID"""
+class BaseAssetIDURL(SingleAssetURL):
+    """
+    Parsed from a URL that refers to an asset by ID and does not include the
+    Dandiset ID
+    """
 
     asset_id: str
 
-    def get_assets(self, client: DandiAPIClient) -> Iterator[RemoteAsset]:
+    def get_assets(self, client: DandiAPIClient) -> Iterator[BaseRemoteAsset]:
+        """
+        Yields the asset with the given ID.  Yields nothing if the asset does
+        not exist.
+        """
+        try:
+            yield client.get_asset(self.asset_id)
+        except requests.HTTPError as e:
+            if e.response.status_code == 404:
+                return
+            else:
+                raise
+
+    def get_asset_ids(self, client: DandiAPIClient) -> Iterator[str]:
+        """Yields the ID of the asset (regardless of whether it exists)"""
+        yield self.asset_id
+
+
+class AssetIDURL(SingleAssetURL):
+    """
+    Parsed from a URL that refers to an asset by ID and includes the Dandiset ID
+    """
+
+    asset_id: str
+
+    def get_assets(self, client: DandiAPIClient) -> Iterator[BaseRemoteAsset]:
         """
         Yields the asset with the given ID.  Yields nothing if the asset does
         not exist.
         """
         try:
             yield self.get_dandiset(client).get_asset(self.asset_id)
-        except requests.HttpError as e:
+        except requests.HTTPError as e:
             if e.response.status_code == 404:
                 return
             else:
@@ -151,9 +187,9 @@ class AssetPathPrefixURL(MultiAssetURL):
     Parsed from a URL that refers to a collection of assets by path prefix
     """
 
-    def get_assets(self, client: DandiAPIClient) -> Iterator[RemoteAsset]:
+    def get_assets(self, client: DandiAPIClient) -> Iterator[BaseRemoteAsset]:
         """Returns the assets whose paths start with `path`"""
-        return self.get_dandiset(client).get_assets_under_path(self.path)
+        return self.get_dandiset(client).get_assets_with_path_prefix(self.path)
 
 
 class AssetItemURL(SingleAssetURL):
@@ -161,7 +197,7 @@ class AssetItemURL(SingleAssetURL):
 
     path: str
 
-    def get_assets(self, client: DandiAPIClient) -> Iterator[RemoteAsset]:
+    def get_assets(self, client: DandiAPIClient) -> Iterator[BaseRemoteAsset]:
         """
         Yields the asset whose path equals `path`.  If there is no such asset,
         this method yields nothing, unless the path happens to be the path to
@@ -173,7 +209,7 @@ class AssetItemURL(SingleAssetURL):
             asset = d.get_asset_by_path(self.path)
         except NotFoundError:
             try:
-                next(d.get_assets_under_path(self.path + "/"))
+                next(d.get_assets_with_path_prefix(self.path + "/"))
             except StopIteration:
                 pass
             else:
@@ -191,7 +227,7 @@ class AssetFolderURL(MultiAssetURL):
 
     path: str
 
-    def get_assets(self, client: DandiAPIClient) -> Iterator[RemoteAsset]:
+    def get_assets(self, client: DandiAPIClient) -> Iterator[BaseRemoteAsset]:
         """
         Returns all assets under the folder at `path`.  Yields nothing if the
         folder does not exist.
@@ -199,22 +235,18 @@ class AssetFolderURL(MultiAssetURL):
         path = self.path
         if not path.endswith("/"):
             path += "/"
-        return self.get_dandiset(client).get_assets_under_path(path)
+        return self.get_dandiset(client).get_assets_with_path_prefix(path)
 
 
 @contextmanager
 def navigate_url(url):
     """Context manager to 'navigate' URL pointing to DANDI archive.
 
-    Parameters
-    ----------
-    url: str
-      URL which might point to a dandiset, a folder, or an asset(s)
+    :param str url: URL which might point to a dandiset, a folder, or an
+        asset(s)
 
-    Yields
-    ------
-    client, dandiset, assets (generator)
-      `client` will have established a session for the duration of the context
+    :returns: Generator of one ``(client, dandiset, assets)``; ``client`` will
+        have established a session for the duration of the context
     """
     parsed_url = parse_dandi_url(url)
     with parsed_url.navigate() as (client, dandiset, assets):
@@ -260,7 +292,7 @@ class _dandi_url_parser:
         (
             re.compile(
                 rf"{server_grp}(#/)?(?P<asset_type>dandiset)/{dandiset_id_grp}"
-                r"(/(?P<version>[.0-9]{5,}|draft))?"
+                rf"(/(?P<version>{VERSION_REGEX}))?"
                 r"(/(files(\?location=(?P<location>.*)?)?)?)?"
             ),
             {},
@@ -278,15 +310,22 @@ class _dandi_url_parser:
         (
             re.compile(
                 rf"{server_grp}(?P<asset_type>dandiset)s/{dandiset_id_grp}"
-                r"(/(versions(/(?P<version>[.0-9]{5,}|draft))?)?)?"
+                rf"(/(versions(/(?P<version>{VERSION_REGEX}))?)?)?"
             ),
             {},
             "https://<server>[/api]/dandisets/<dandiset id>[/versions[/<version>]]",
         ),
+        (
+            re.compile(
+                rf"{server_grp}(?P<asset_type>asset)s/(?P<asset_id>[^?/]+)(/(download/?)?)?"
+            ),
+            {},
+            "https://<server>[/api]/assets/<asset id>[/download]",
+        ),
         (
             re.compile(
                 rf"{server_grp}(?P<asset_type>dandiset)s/{dandiset_id_grp}"
-                r"/versions/(?P<version>[.0-9]{5,}|draft)"
+                rf"/versions/(?P<version>{VERSION_REGEX})"
                 r"/assets/(?P<asset_id>[^?/]+)(/(download/?)?)?"
             ),
             {},
@@ -296,7 +335,7 @@ class _dandi_url_parser:
         (
             re.compile(
                 rf"{server_grp}(?P<asset_type>dandiset)s/{dandiset_id_grp}"
-                r"/versions/(?P<version>[.0-9]{5,}|draft)"
+                rf"/versions/(?P<version>{VERSION_REGEX})"
                 r"/assets/\?path=(?P<path>[^&]+)",
             ),
             {},
@@ -312,7 +351,7 @@ class _dandi_url_parser:
             re.compile(
                 rf"dandi://(?P<instance_name>({'|'.join(known_instances)}))"
                 rf"/{dandiset_id_grp}"
-                r"(@(?P<version>[.0-9]{5,}|draft))?"
+                rf"(@(?P<version>{VERSION_REGEX}))?"
                 rf"(/(?P<location>.*)?)?"
             ),
             {},
@@ -337,14 +376,17 @@ class _dandi_url_parser:
 
     @classmethod
     def parse(cls, url, *, map_instance=True):
-        """Parse url like and return server (address), asset_id and/or directory
+        """
+        Parse url like and return server (address), asset_id and/or directory
 
         Example URLs (as of 20210428):
+
         - Dataset landing page metadata:
           https://gui.dandiarchive.org/#/dandiset/000003
 
         Individual and multiple files:
-          - dandi???
+
+        - dandi???
 
         Multiple selected files + folders -- we do not support ATM, then further
         RFing would be due, probably making this into a generator or returning a
@@ -352,12 +394,10 @@ class _dandi_url_parser:
 
         "Features":
 
-        - uses some of `known_instance`s to map some urls, e.g. from
+        - uses some of `known_instances` to map some urls, e.g. from
           gui.dandiarchive.org ones into girder.
 
-        Returns
-        -------
-        ParsedDandiURL
+        :rtype: ParsedDandiURL
         """
         lgr.debug("Parsing url %s", url)
 
@@ -421,8 +461,13 @@ class _dandi_url_parser:
             )
 
         url_server = groups["server"].rstrip("/")
+        if re.fullmatch(
+            r"https?://deploy-preview-.*--gui-dandiarchive-org\.netlify\.app",
+            url_server,
+        ):
+            url_server = "https://gui-staging.dandiarchive.org"
         server = cls.map_to.get(url_server, url_server)
-        asset_type = groups.get("asset_type")
+        # asset_type = groups.get("asset_type")
         dandiset_id = groups.get("dandiset_id")
         version_id = groups.get("version")
         location = groups.get("location")
@@ -433,8 +478,6 @@ class _dandi_url_parser:
             # ATM carries leading '/' which IMHO is not needed/misguiding somewhat, so
             # I will just strip it
             location = location.lstrip("/")
-        if not (asset_type == "dandiset" and dandiset_id):
-            raise ValueError(f"{url} does not point to a dandiset")
 
         # if location is not degenerate -- it would be a folder or a file
         if location:
@@ -453,12 +496,15 @@ class _dandi_url_parser:
                     path=location,
                 )
         elif asset_id:
-            parsed_url = AssetIDURL(
-                api_url=server,
-                dandiset_id=dandiset_id,
-                version_id=version_id,
-                asset_id=asset_id,
-            )
+            if dandiset_id is None:
+                parsed_url = BaseAssetIDURL(api_url=server, asset_id=asset_id)
+            else:
+                parsed_url = AssetIDURL(
+                    api_url=server,
+                    dandiset_id=dandiset_id,
+                    version_id=version_id,
+                    asset_id=asset_id,
+                )
         elif path:
             parsed_url = AssetPathPrefixURL(
                 api_url=server,
@@ -477,18 +523,24 @@ class _dandi_url_parser:
 
     @staticmethod
     def follow_redirect(url):
-        r = requests.head(url, allow_redirects=True)
-        if r.status_code == 404:
-            raise NotFoundError(url)
-        elif r.status_code != 200:
-            raise FailedToConnectError(
-                f"Response for getting {url} to redirect returned {r.status_code}."
-                f" Please verify that it is a URL related to dandiarchive and"
-                f" supported by dandi client"
-            )
-        elif r.url != url:
-            return r.url
-        return url
+        i = 0
+        while True:
+            r = requests.head(url, allow_redirects=True)
+            if r.status_code in (400, 502, 503, 504) and i < 5:
+                sleep(0.1 * 10 ** i)
+                i += 1
+                continue
+            elif r.status_code == 404:
+                raise NotFoundError(url)
+            elif r.status_code != 200:
+                raise FailedToConnectError(
+                    f"Response for getting {url} to redirect returned {r.status_code}."
+                    f" Please verify that it is a URL related to dandiarchive and"
+                    f" supported by dandi client"
+                )
+            elif r.url != url:
+                return r.url
+            return url
 
 
 # convenience binding
